/*
 * Copyright (c) 2017 N. Billard.
 */

package com.massa.mrules.extensions.dsl.editor.web;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.vaadin.aceeditor.AceEditor;
import org.vaadin.aceeditor.SuggestionExtension;

import com.massa.dsl.doc.DslDocGen;
import com.massa.dsl.grammar.Grammar;
import com.massa.dsl.grammar.GrammarRepository;
import com.massa.mrules.extensions.dsl.doc.html.DslHtmlDocGen;
import com.massa.util.StringUtils;
import com.massa.util.UtilsException;
import com.massa.util.classloader.ResourceLoader;
import com.vaadin.annotations.JavaScript;
import com.vaadin.annotations.Theme;
import com.vaadin.data.HasValue.ValueChangeEvent;
import com.vaadin.data.HasValue.ValueChangeListener;
import com.vaadin.server.StreamResource;
import com.vaadin.server.StreamResource.StreamSource;
import com.vaadin.server.VaadinRequest;
import com.vaadin.spring.annotation.SpringUI;
import com.vaadin.ui.Button;
import com.vaadin.ui.HorizontalLayout;
import com.vaadin.ui.NativeSelect;
import com.vaadin.ui.Notification;
import com.vaadin.ui.UI;
import com.vaadin.ui.VerticalLayout;

/**
 * Main UI for DSL editor.
 */
@SpringUI
@Theme("mrules_theme")
@JavaScript({ "vaadin://js/filesaver.js" })
@JavaScript({ "vaadin://js/save.js" })
public class MRulesDslEditorUI extends UI {

	private static final class DocGenStreamSource implements StreamSource {
		private final Grammar g;
		private static final long serialVersionUID = 1L;

		private DocGenStreamSource(final Grammar g) {
			this.g = g;
		}

		@Override
		public InputStream getStream() {
			try {
				final ByteArrayOutputStream baos = new ByteArrayOutputStream(5*1024);
				final DslDocGen gen = new DslHtmlDocGen(this.g);
				gen.generate(baos);
				return new ByteArrayInputStream(baos.toByteArray());
			} catch (final UtilsException e) {
				throw new RuntimeException(e);
			}
		}
	}

	private static final long serialVersionUID = 1L;

	private final AceEditor editor;

	private final NativeSelect<String> grammars;
	private final Button saveBtn;
	private final Button resetBtn;
	private final Button docBtn;

	@Value("${suggestion.extension.class:#{null}}")
	private String suggestionExtensionClass;

	@Value("${grammar.doc.link.visible:#{true}}")
	private boolean grammarDocLink;

	private SuggestionExtension suggestionExtension;

	@Autowired
	public MRulesDslEditorUI() {
		this.editor = new AceEditor();
		this.grammars = new NativeSelect<>();
		this.saveBtn = new Button("Save");
		this.resetBtn = new Button("Reset");
		this.docBtn = new Button("Generate Documentation");
	}

	@Override
	protected void init(final VaadinRequest request) {
		// build layout
		final HorizontalLayout actions;
		if (this.grammarDocLink) {
			actions = new HorizontalLayout(this.grammars, this.saveBtn, this.resetBtn, this.docBtn);
		} else {
			actions = new HorizontalLayout(this.grammars, this.saveBtn, this.resetBtn);
		}

		this.editor.setWidth(90, Unit.PERCENTAGE);
		this.editor.setHeight(90, Unit.PERCENTAGE);
		this.editor.setId("mrules_editor");

		final VerticalLayout mainLayout = new VerticalLayout(actions, this.editor);
		mainLayout.setExpandRatio(actions, 1);
		mainLayout.setExpandRatio(this.editor, 15);
		mainLayout.setWidth(100, Unit.PERCENTAGE);
		mainLayout.setHeight(100, Unit.PERCENTAGE);
		this.setContent(mainLayout);

		final List<String> grammarIds = GrammarRepository.listGrammarIds();

		if (grammarIds.isEmpty()) {
			final List<String> grammarItems = new ArrayList<>();
			grammarItems.add("No loaded grammar.");
			this.grammars.setItems(grammarItems);
			this.grammars.setSelectedItem("No loaded grammar.");
			this.grammars.setEnabled(false);
			this.editor.setEnabled(false);
			this.saveBtn.setEnabled(false);
			this.resetBtn.setEnabled(false);
			this.docBtn.setEnabled(false);
		} else {
			this.saveBtn.addClickListener(e -> this.save());
			this.resetBtn.addClickListener(e -> this.reset());
			this.docBtn.addClickListener(e -> this.genDoc());
			final MRulesDslSuggester mRulesDslSuggester;

			if (this.suggestionExtensionClass == null || StringUtils.isEmpty(this.suggestionExtensionClass)) {
				mRulesDslSuggester = new MRulesDslSuggester(this.editor);
			} else {
				try {
					final Class<?> clazz = ResourceLoader.loadClass(this.suggestionExtensionClass);
					mRulesDslSuggester = (MRulesDslSuggester) clazz.getConstructor(AceEditor.class).newInstance(this.editor);
				} catch (ClassNotFoundException | NoSuchMethodException | SecurityException | InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
					throw new RuntimeException(e);
				}
			}
			this.suggestionExtension = new SuggestionExtension(mRulesDslSuggester);

			this.suggestionExtension = new SuggestionExtension(mRulesDslSuggester);
			this.suggestionExtension.extend(this.editor);
			this.suggestionExtension.setShowDescriptions(true);
			this.suggestionExtension.setSuggestOnDot(false);
			this.suggestionExtension.setPopupWidth(100, Unit.PERCENTAGE);
			this.suggestionExtension.setPopupHeight(100, Unit.PERCENTAGE);
			this.suggestionExtension.setpopupDescriptionWidth(100, Unit.PERCENTAGE);

			// Listen changes made by the editor, refresh data from backend
			this.editor.addValueChangeListener(mRulesDslSuggester);

			final List<String> grammarItems = new ArrayList<>();
			boolean first = true;
			for (final String grammar: grammarIds) {
				final List<String> locales = GrammarRepository.listGrammarLocales(grammar);
				for (final String locale: locales) {
					grammarItems.add(grammar + " - " + locale);
					if (first) {
						((MRulesDslSuggester) MRulesDslEditorUI.this.suggestionExtension.getSuggester()).setGrammar(
								GrammarRepository.getFromId(grammar, locale));
						first = false;
					}
				}
			}
			this.grammars.setItems(grammarItems);
			if (!grammarItems.isEmpty()) {
				this.grammars.setSelectedItem(grammarItems.get(0));
			}
			this.grammars.addValueChangeListener(new ValueChangeListener<String>() {
				private static final long serialVersionUID = 1L;
				@Override
				public void valueChange(final ValueChangeEvent<String> event) {
					if (event.getValue() != null && !event.getValue().isEmpty()) {
						final String[] grammarAndLocale = event.getValue().split(" - ");
						((MRulesDslSuggester) MRulesDslEditorUI.this.suggestionExtension.getSuggester()).setGrammar(
								GrammarRepository.getFromId(grammarAndLocale[0], grammarAndLocale[1]));
					}
				}
			});
		}
	}

	void save() {
		com.vaadin.ui.JavaScript.getCurrent().execute("saveRuleset()");
		Notification.show("Saved", "Saved", Notification.Type.HUMANIZED_MESSAGE);
	}

	void reset() {
		this.editor.setValue("");
	}

	@SuppressWarnings("deprecation")
	void genDoc() {
		final Grammar g = ((MRulesDslSuggester) MRulesDslEditorUI.this.suggestionExtension.getSuggester()).getGrammar();
		final String gid = g.getId();
		final StreamSource source = new DocGenStreamSource(g);
		final StreamResource resource = new StreamResource(source, "doc/" + gid);
		resource.setMIMEType("text/html");
		this.getPage().open(resource, "doc/" + gid, true);
	}

}
