/*
 * Copyright (c) 2017 N. Billard.
 */

package com.massa.mrules.extensions.dsl.editor.web;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.vaadin.aceeditor.AceEditor;
import org.vaadin.aceeditor.Suggester;
import org.vaadin.aceeditor.Suggestion;
import org.vaadin.aceeditor.TextRange;
import org.vaadin.aceeditor.client.AceAnnotation;
import org.vaadin.aceeditor.client.AceMarker;
import org.vaadin.aceeditor.client.AceRange;

import com.massa.dsl.DslException;
import com.massa.dsl.grammar.Grammar;
import com.massa.dsl.grammar.ParsingContext.AutoCompleteResult;
import com.massa.util.AdvancedReader;
import com.vaadin.data.HasValue.ValueChangeEvent;
import com.vaadin.data.HasValue.ValueChangeListener;

/**
 * Adds feature to the defaut editor suggester. Also handles syntax check after text changes.<br>
 * This implementation can be overridden to add features using the "suggestion.extension.class" application property.
 */
public class MRulesDslSuggester implements Suggester, ValueChangeListener<String> {
	private static final long serialVersionUID = 1L;

	private final AceEditor editor;
	private Grammar grammar;

	/**
	 * Adds data to a suggestion to be more accurate.
	 */
	public static class MySuggestion extends Suggestion {
		private final String insertThis;
		MySuggestion(final String group, final String displayText, final String descriptionText, final String suggestionText, final Integer suggestionStart, final Integer suggestionEnd) {
			super(group, displayText, descriptionText, suggestionText, suggestionStart, suggestionEnd);
			this.insertThis = suggestionText;
		}
	}

	public MRulesDslSuggester(final AceEditor editor) {
		this.editor = editor;
	}

	@Override
	public String applySuggestion(final Suggestion s, final String text, final int cursor) {
		final String ins = ((MySuggestion) s).insertThis;
		final TextRange selection = this.editor.getSelection();
		final int start = selection.getStart() < selection.getEnd() ? selection.getStart() : selection.getEnd();
		final int end = selection.getStart() > selection.getEnd() ? selection.getStart() : selection.getEnd();
		return text.substring(0, start) + ins + text.substring(end);
	}

	@Override
	public List<Suggestion> getSuggestions(final String text, final int cursor) {
		final List<Suggestion> results = new ArrayList<>();
		if (this.grammar != null) {
			try {
				final TextRange selection = this.editor.getSelection();
				final AdvancedReader source = new AdvancedReader(new StringReader(text));
				source.setMaxReadLen(selection.getStart() < selection.getEnd() ? selection.getStart() : selection.getEnd());
				final Set<AutoCompleteResult> autocomps = this.grammar.autoComplete(source);
				for (final AutoCompleteResult autocomp: autocomps) {
					final MySuggestion result = new MySuggestion(
							autocomp.getGroup(),
							autocomp.getAutoCompleteLabel(),
							autocomp.getAutoComplete().replaceAll("\n", "<br/>"),
							autocomp.getAutoComplete().substring(autocomp.getStartingPoint()),
							autocomp.getSelectionStart() == null ? null : autocomp.getSelectionStart() - autocomp.getStartingPoint(),
							autocomp.getSelectionEnd() == null ? null : autocomp.getSelectionEnd() - autocomp.getStartingPoint());
					results.add(result);
				}
			} catch (final DslException e) {
				//Silence
			}
		}
		return results;
	}

	/**
	 * Validates syntax after value changes. Creates marks if necessary.
	 */
	@Override
	public void valueChange(final ValueChangeEvent<String> event) {
		this.editor.clearRowAnnotations();
		this.editor.clearMarkers();
		try {
			this.grammar.validate(new AdvancedReader(new StringReader(event.getValue())));
		} catch (final DslException e) {
			final int line = e.getLine() == null ? 0 : e.getLine();
			final int column = e.getColumn() == null ? 0 : e.getColumn();

			final AceMarker marker = new AceMarker("XYZ", new AceRange(line - 1, column - 1, line - 1, column), "mruleserror", AceMarker.Type.text, false, AceMarker.OnTextChange.ADJUST);
			this.editor.addMarker(marker);
			this.editor.addRowAnnotation(new AceAnnotation(e.getMessageInfo().getMessage(), AceAnnotation.Type.error), line - 1);
		}
	}

	public AceEditor getEditor() {
		return this.editor;
	}

	public Grammar getGrammar() {
		return this.grammar;
	}

	public void setGrammar(final Grammar grammar) {
		this.grammar = grammar;
	}
}
